/* -*- c++ -*- */
/////////////////////////////////////////////////////////////////////////////
//
// Particles.cpp  -- Copyright (c) 2006 David Henry
// last modification: feb. 9, 2006
//
// This code is licenced under the MIT license.
//
// This software is provided "as is" without express or implied
// warranties. You may freely copy and compile this source into
// applications you distribute provided that the copyright text
// below is included in the resulting source code.
//
// Implementation of particle systems related classes.
//
/////////////////////////////////////////////////////////////////////////////

#include <GL/glew.h>
#include <iostream>
#include <cmath>

#include "Particles.h"

extern bool bUseExtensions;


/////////////////////////////////////////////////////////////////////////////
//
// class Particle implementation.
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
// Particle::Particle
//
// Constructor.  The particle is not yet alive; it will be when
// emitted by a particle system.
// --------------------------------------------------------------------------

Particle::Particle ()
  : _life (0.0f), _alive (false), _size (1.0f), _weight (0.0f)
{
}


// --------------------------------------------------------------------------
// Particle::~Particle
//
// Destructor.
// --------------------------------------------------------------------------

Particle::~Particle ()
{
}


// --------------------------------------------------------------------------
// Particle::init
//
// Initialize the particle.  The particle will now be alive.
// --------------------------------------------------------------------------

void
Particle:: init (const vec3_t &pos, const vec3_t &vel, const vec3_t &accel,
		 float life, float size, float weight, const vec4_t &color)
{
  memcpy (_position, pos, sizeof (vec3_t));
  memcpy (_velocity, vel, sizeof (vec3_t));
  memcpy (_acceleration, accel, sizeof (vec3_t));
  memcpy (_color, color, sizeof (vec4_t));

  _life = life;
  _alive = true;
  _size = size;
  _weight = weight;
}


// --------------------------------------------------------------------------
// Particle::update
//
// Update particle's state.
// --------------------------------------------------------------------------

void
Particle::update (float dt)
{
  if (_alive)
    {
      _life -= dt;

      // Is the particle still alive?
      if (_life <= 0.0)
	{
	  _alive = false;
	}
      else
	{
	  // Compute particle's velocity and position
	  _velocity[0] += _acceleration[0] * dt;
	  _velocity[1] += _acceleration[1] * dt;
	  _velocity[2] += _acceleration[2] * dt;

	  _position[0] += _velocity[0] * dt;
	  _position[1] += _velocity[1] * dt;
	  _position[2] += _velocity[2] * dt;

	  // Basic collision detection
	  if (_velocity[1] < 0.0f)
	    {
	      // Bounce on the floor
	      if (_position[1] < (_size * 0.5f))
		_velocity[1] = -kFriction * _velocity[1] + ((rand() % 255) / 255.0f);
	    }
	}
    }
}


// --------------------------------------------------------------------------
// Particle::drawPoint
//
// Draw the particle as a colored point.
// --------------------------------------------------------------------------

void
Particle::drawPoint (bool reflection)
{
  if (reflection)
    {
      // Apply linear attenuation if rendering the
      // reflection of the particle
      glColor4f (_color[0], _color[1], _color[2],
		 _color[3] - (_position[1] * 0.2f));
    }
  else
    {
      glColor4fv (_color);
    }

  glVertex3fv (_position);
}


// --------------------------------------------------------------------------
// Particle::drawQuad
//
// Draw the particle as a quad, using bilboarding technique given
// right and up vectors from the modelview matrix.  Assume that the
// glBegin()/glEnd() functions are called outside this function.
// --------------------------------------------------------------------------

void
Particle::drawQuad (const vec3_t &right, const vec3_t &up, bool reflection)
{
  if (reflection)
    {
      // Apply linear attenuation if rendering the
      // reflection of the particle
      glColor4f (_color[0], _color[1], _color[2],
		 _color[3] - (_position[1] * 0.2f));
    }
  else
    {
      glColor4fv (_color);
    }

  // Draw the quad
  glTexCoord2f (0.0f, 0.0f);
  glVertex3f (_position[0] - (right[0] + up[0]) * _size,
	      _position[1] - (right[1] + up[1]) * _size,
	      _position[2] - (right[2] + up[2]) * _size );

  glTexCoord2f (1.0f, 0.0f);
  glVertex3f (_position[0] + (right[0] - up[0]) * _size,
	      _position[1] + (right[1] - up[1]) * _size,
	      _position[2] + (right[2] - up[2]) * _size);

  glTexCoord2f (1.0f, 1.0f);
  glVertex3f (_position[0] + (right[0] + up[0]) * _size,
	      _position[1] + (right[1] + up[1]) * _size,
	      _position[2] + (right[2] + up[2]) * _size);

  glTexCoord2f (0.0f, 1.0f);
  glVertex3f (_position[0] + (up[0] - right[0]) * _size,
	      _position[1] + (up[1] - right[1]) * _size,
	      _position[2] + (up[2] - right[2]) * _size);
}


// --------------------------------------------------------------------------
// Particle::drawLight
//
// Draw the particle's light on the floor as a lightmap.  Assume that
// the glBegin()/glEnd() functions are called with GL_QUADS outside
// this function.
// --------------------------------------------------------------------------

void
Particle::drawLight ()
{
  if (_position[1] < 1.0)
    {
      // White light with linear attenuation
      glColor4f (1.0, 1.0, 1.0, (1.0 - _position[1]) * 0.3);
      float size = _size * 3;

      // Draw the lightmap as a quad
      glTexCoord2f (0.0f, 1.0f);
      glVertex3f (_position[0] - size, 0.01f, _position[2] + size);

      glTexCoord2f (1.0f, 1.0f);
      glVertex3f (_position[0] + size, 0.01f, _position[2] + size);

      glTexCoord2f (1.0f, 0.0f);
      glVertex3f (_position[0] + size, 0.01f, _position[2] - size);

      glTexCoord2f (0.0f, 0.0f);
      glVertex3f (_position[0] - size, 0.01f, _position[2] - size);
    }
}


/////////////////////////////////////////////////////////////////////////////
//
// class ParticleSystem implementation.
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
// ParticleSystem::ParticleSystem
//
// Constructor.  Initialize the particle system.  Create a list of
// pre-built particles.
// --------------------------------------------------------------------------

ParticleSystem::ParticleSystem (int maxParticles, float life,
				const vec3_t &origin)
  : _emitting (false), _maxParticles (maxParticles), _numParticles (0),
    _particleLife (life), _accumulatedTime (0.0f)
{
  _origin[0] = origin[0];
  _origin[1] = origin[1];
  _origin[2] = origin[2];

  _particleList = new Particle[_maxParticles];
}


// --------------------------------------------------------------------------
// ParticleSystem::~ParticleSystem
//
// Destructor.  Destroy the particles.
// --------------------------------------------------------------------------

ParticleSystem::~ParticleSystem ()
{
  delete [] _particleList;
}


// --------------------------------------------------------------------------
// ParticleSystem::emit
//
// Emit a given number of particles.  Return the number of particles
// really emitted.
// --------------------------------------------------------------------------

int
ParticleSystem::emit (int numParticles)
{
  while (numParticles && (numParticles < _maxParticles))
    {
      initializeParticle (_particleList[_numParticles]);
      _numParticles++;
      numParticles--;
    }

  return numParticles;
}


/////////////////////////////////////////////////////////////////////////////
//
// class ParticleFountain implementation.
//
/////////////////////////////////////////////////////////////////////////////

// --------------------------------------------------------------------------
// ParticleFountain::ParticleFountain
//
// Constructor.  Initialize the particle fountain.
// --------------------------------------------------------------------------

ParticleFountain::ParticleFountain (int maxParticles, float life,
			    const vec3_t &origin, const vec3_t &emitDir)
  : ParticleSystem (maxParticles, life, origin),
    _emit (0.0f), _texId (0)
{
  _emitDir[0] = emitDir[0];
  _emitDir[1] = emitDir[1];
  _emitDir[2] = emitDir[2];
}


// --------------------------------------------------------------------------
// ParticleFountain::update
//
// Update the particle system and all owned particles.
// --------------------------------------------------------------------------

void
ParticleFountain::update (float dt)
{
  // Update time from particle system creation
  _accumulatedTime += dt;

  // Update each particle
  for (int i = 0; i < _maxParticles; ++i)
    {
      _particleList[i].update (dt);

      if (!_particleList[i].isAlive ())
	_numParticles--;
    }
}


// --------------------------------------------------------------------------
// ParticleFountain::render
//
// Draw the particles.  If reflection is set, assume we are drawing
// the reflection of the particles on the floor.
// --------------------------------------------------------------------------

void
ParticleFountain::render (bool reflection)
{
  // Bind to particles' texture
  glEnable (GL_TEXTURE_2D);
  glBindTexture (GL_TEXTURE_2D, _texId);

  // Enable alpha blending and disable depth buffer.
  glEnable (GL_BLEND);
  glBlendFunc (GL_SRC_ALPHA, GL_ONE);
  glDepthMask (GL_FALSE);

  // Draw each particle
  if (GLEW_ARB_point_sprite && GLEW_ARB_point_parameters && bUseExtensions)
    {
      //
      // Draw particles as points using the GL_POINT_SPRITE_ARB
      // extension.
      //

      // Setup point sprite extension
      glTexEnvf (GL_POINT_SPRITE_ARB, GL_COORD_REPLACE_ARB, GL_TRUE);
      glEnable (GL_POINT_SPRITE_ARB);

      static float quadratic[] = { 0.0f, 0.0f, 0.01f };
      glPointParameterfv (GL_POINT_DISTANCE_ATTENUATION, quadratic);

      float fMaxSize = 0.0f;
      glGetFloatv (GL_POINT_SIZE_MAX, &fMaxSize);

      glPointSize (fMaxSize);
      glPointParameterf (GL_POINT_SIZE_MAX, fMaxSize);
      glPointParameterf (GL_POINT_SIZE_MIN, 0.1f);

      // Draw the particles as points
      glBegin (GL_POINTS);
	for (int i = 0; i < _maxParticles; ++i)
	  {
	    if (_particleList[i].isAlive ())
	      _particleList[i].drawPoint (reflection);
	  }
      glEnd();

      glDisable (GL_POINT_SPRITE_ARB);
      glTexEnvf (GL_POINT_SPRITE_ARB, GL_COORD_REPLACE_ARB, GL_FALSE);
    }
  else
    {
      //
      // Draw particles as textured quads.
      //

      // Get modelview matrix and extract right and up
      // vectors for bilbioarding
      GLfloat mat[16];
      glGetFloatv (GL_MODELVIEW_MATRIX, mat);

      vec3_t right = { mat[0], mat[4], mat[8] };
      vec3_t up    = { mat[1], mat[5], mat[9] };

      // Draw the particles as quads
      glBegin (GL_QUADS);
	for (int i = 0; i < _maxParticles; ++i)
	  {
	    if (_particleList[i].isAlive ())
	      _particleList[i].drawQuad (right, up, reflection);
	  }
      glEnd ();
    }

  glDepthMask (GL_TRUE);

  // les particules près du sol sont source de lumière
  // on dessine donc des lightmaps sur le sol lorsque
  // les particules en sont assez proches

  // Particles which are near the floor emit light
  // on it.  Draw the lightmaps on the floor.
  if (!reflection)
    {
      // We don't draw lightmaps if we are in reflection
      // rendering mode

      glBegin (GL_QUADS);
	for (int i = 0; i < _maxParticles; ++i)
	  {
	    if (_particleList[i].isAlive ())
	      _particleList[i].drawLight ();
	  }
      glEnd ();
    }

  glDisable (GL_BLEND);
  glDisable (GL_TEXTURE_2D);
}


// --------------------------------------------------------------------------
// ParticleFountain::emit
//
// Emit a given number of particles.  This number can be a floating
// point number: an accumulator will store the number of particles to
// emit, and emit ones when greater or equal than zero.
// Return the number of really emitted particles.
// --------------------------------------------------------------------------

int
ParticleFountain::emit (float fNumParticles)
{
  // If we are not in emit mode, exit.
  if (!_emitting)
    return 0;

  // Update accumulator
  _emit += fNumParticles;

  int numParticles = 0;
  int emited = 0;
  int lastFreeSlot = 0;

  // Can we emit one or more particles?
  if (_emit > 1.0)
    {
      // Get the entire number of particles we
      // can emit.
      numParticles = static_cast<int>(_emit);
      _emit -= numParticles;
    }

  // Emit as many particles as we can
  while (numParticles && (_numParticles < _maxParticles))
    {
      // Look for a freeslot (a dead particle in the list)
      for (int i = lastFreeSlot; i < _maxParticles; ++i)
	{
	  if (!_particleList[i].isAlive ())
	    {
	      // We can reuse this particle,
	      // reinit it.
	      initializeParticle (_particleList[i]);

	      _numParticles++;
	      numParticles--;
	      emited++;

	      // Update the next start index for searching
	      // a free slot in the list
	      lastFreeSlot = i + 1;
	      break;
	    }
	}
    }

  return emited;
}


// --------------------------------------------------------------------------
// ParticleFountain::initializeParticle
//
// Initialize a particle and emit it.
// --------------------------------------------------------------------------

void ParticleFountain::initializeParticle (Particle &p)
{
  float t = _accumulatedTime;

  // Compute velocity
  vec3_t velocity;
  float theta = (static_cast<float>(rand() % 360) - 180) * 180 / kPi;
  velocity[0] = _emitDir[0] + std::cos (theta) * 2.5f;
  velocity[1] = _emitDir[1] + static_cast<float>(rand() % 4095) * 0.000025f * _emitDir[1];
  velocity[2] = _emitDir[2] + std::sin (theta) * 2.5f;

  // Acceleration is constant
  vec3_t acceleration;
  acceleration[0] = 0.0f;
  acceleration[1] = -kGravity;
  acceleration[2] = 0.0f;

  // Pick a cool color for the particle, in function of time
  vec4_t color;
  color[0] = 0.7f + 0.3f * std::sin (0.34 * t + 0.1);
  color[1] = 0.6f + 0.4f * std::sin (0.63 * t + 1.1);
  color[2] = 0.6f + 0.4f * std::sin (0.97 * t + 2.1);
  color[3] = 1.0;

  p.init (_origin, velocity, acceleration, _particleLife, 0.5f, 1.0f, color);
}
